# imdb_lstm.py

import numpy as np
import cntk as C

# def mb_accuracy(mb, x_var, y_var, model):
#   num_correct = 0; num_wrong = 0
# 
#   xmat = mb[x_var]
#   ymat = mb[y_var].as_sequences(y_var) # targets like [[0, 1]]
# 
#   model = C.ops.softmax(model)  # make computed outpt sum to 1.0
#   computed = model.eval(xmat)  # like [0, 1]
#   for i in range(len(ymat)):
#     if np.argmax(ymat[i][0]) == np.argmax(computed[i]):
#       num_correct += 1
#     else:
#       num_wrong += 1
#   acc = (num_correct * 100) / (num_correct + num_wrong)
#   return acc

def create_reader(path, input_dim, output_dim, is_random, sweeps):
  x_strm = C.io.StreamDef(field='x', shape=input_dim, is_sparse=True)
  y_strm = C.io.StreamDef(field='y', shape=output_dim, is_sparse=False)
  streams = C.io.StreamDefs(x_src=x_strm, y_src=y_strm)
  deserial = C.io.CTFDeserializer(path, streams)
  mb_source = C.io.MinibatchSource(deserial, randomize=is_random, max_sweeps=sweeps)
  return mb_source

def main():
  # 0. get started
  print("\nBegin IMDB sentiment analysis with CNTK LSTM \n")
  np.random.seed(1)

  train_file = ".\\Data\\imdb_sparse_train_50w.txt"
  test_file = ".\\Data\\imdb_sparse_test_50w.txt"

  input_dim = 129888 + 4  # number distinct input words + offset for one-hot, sparse
  output_dim = 2  # number classes (binary) 

  X = C.sequence.input_variable(shape=input_dim, is_sparse=True)  # sequences
  Y = C.ops.input_variable(output_dim)

  # 1. prepare data reader
  rdr = create_reader(train_file, input_dim, output_dim,
    is_random=True, sweeps=C.io.INFINITELY_REPEAT)

  imdb_map = {
    X : rdr.streams.x_src,  # see create_reader()
    Y : rdr.streams.y_src
  }

  # 2. define LSM model
  my_init = C.initializer.glorot_uniform(seed=1)
  lstm = C.layers.Sequential([
    C.layers.Embedding(50, init=my_init),
    C.layers.Recurrence(C.layers.LSTM(25)),
    C.sequence.last,
    C.layers.Dense(output_dim, init=my_init)])
  model = lstm(X)
  # model = C.ops.softmax(model)

  # 3. no compilation required

  # 4. train model
  max_iter = 400 
  batch_size = 10 * 40  # ~9 sequences
  lr = C.learning_parameter_schedule_per_sample(0.1)
  sgd = C.sgd(model.parameters, lr)
  tr_loss = C.cross_entropy_with_softmax(model, Y)
  tr_accu = C.classification_error(model, Y)
  trainer = C.Trainer(model, (tr_loss,tr_accu), [sgd]) 

  print("Training model . . . . \n")
  for i in range(max_iter):
    mb = rdr.next_minibatch(batch_size, input_map=imdb_map)
    trainer.train_minibatch(mb)

    if i % int(max_iter/5) == 0:
      print("i = %d" % i)
      num_seqs = mb[Y].num_sequences
      print("curr mini-batch has %d sequences" % num_seqs)

      curr_class_err = (1.0 - trainer.previous_minibatch_evaluation_average) * 100
      print("accuracy curr mb = %0.2f%%" % curr_class_err) 

      curr_loss = trainer.previous_minibatch_loss_average
      print("loss curr mb = %0.4f" % curr_loss)
 
      print("----------")    
  print("\nTraining complete \n") 

  # 5. evaluate model
  print("Evaluating model on test items ")
  rdr = create_reader(test_file, input_dim, output_dim,
    is_random=False, sweeps=1)

  imdb_map = {
    X : rdr.streams.x_src,  # see create_reader()
    Y : rdr.streams.y_src
  }

  num_to_test = 500 * 25000  # guaranteed to catch all items if reviews len <= 500
  all_test = rdr.next_minibatch(num_to_test, input_map=imdb_map) 
  class_acc = (1.0 - trainer.test_minibatch(all_test)) * 100
  print("Classification accuracy on all test items = %0.2f%%" % class_acc) 

  # acc = mb_accuracy(all_test, X, Y, model)
  # print("\nAccuracy computed by custom function = %0.2f" % acc)

  # 6. save model
  # mp = ".\\Models\\imdb_model.cntk"
  # model.save(mp)

  # 7. make a prediction
  print("Predicting sentiment for \'i like this movie\'")
  print("where [1 0] is negative and [0 1] is positive")
  review_file = ".\\Data\\my_reviews_50w.txt"
  rdr = create_reader(review_file, input_dim, output_dim,
    is_random=False, sweeps=1)

  imdb_map = {
    X : rdr.streams.x_src,  # see create_reader()
    Y : rdr.streams.y_src
  }
  review_mb = rdr.next_minibatch(1, input_map=imdb_map)
  model = C.ops.softmax(model)
  predicted = model.eval(review_mb[X])
  print(predicted)

if __name__ == "__main__":
  main()
   